#ifndef FGI_SCENERY_H
#define FGI_SCENERY_H

#include <wx/wx.h>
#include <wx/filename.h>
#include <wx/dynarray.h>

#include "FGITile.h"
#include "FGIModel.h"

WX_DECLARE_OBJARRAY(FGITile, FGITileArray);
WX_DECLARE_OBJARRAY(FGIModel, FGIModelArray);

enum FGISceneryType
{
    SceneryUnknown = 0x0000,
    SceneryXPDsf   = 0x0001,
    SceneryXPEnv   = 0x0002
};

class FGIScenery
{
private:
    wxString       Name;
    wxFileName     Path;
    FGISceneryType Type;
    wxFileName     OutputDir;

    wxInt32        CurrentModel;
    void           CleanIndexes();

    bool           StaticModels;
    bool           GroupModels;

public:
    FGIScenery(wxString name, wxFileName path, wxFileName output, bool static_mod, bool group_mod);

    bool           Found();
    bool           Load();

    wxString       GetName();
    bool           IsStatic();

    FGITileArray   Tiles;
    FGITile*       GetTile(wxString name, wxString path);

    FGIModelArray  Models;
    wxInt32        GetModelFromXPIndex(wxInt32 xpindex);

    /* DSF related functions */
    static bool    DSFNextPass(int finished_pass_index, void * inRef);
    static void    DSFAcceptObjectDef(const char * inPartialPath, void * inRef);
    static void    DSFAddObject(unsigned int inObjectType, double inCoordinates[2], double inRotation, void * inRef);
    bool           LoadDSF(wxString name);
};

#endif

