#include <wx/wx.h>
#include <wx/filename.h>
#include <math.h>

#include "FGITile.h"

#include <wx/arrimpl.cpp>
WX_DEFINE_OBJARRAY(FGIObjectArray);

FGITile::FGITile(wxString name, wxString path)
{
    Name = name;
    Path = path;
}

wxDouble FGITile::BucketSpan(wxDouble lat)
{
    if (lat >= 89.0 )
        return 360.0;
    else if (lat >= 88.0 )
        return 8.0;
    else if (lat >= 86.0 )
        return 4.0;
    else if (lat >= 83.0 )
        return 2.0;
    else if (lat >= 76.0 )
        return 1.0;
    else if (lat >= 62.0 )
        return 0.5;
    else if (lat >= 22.0 )
        return 0.25;
    else if (lat >= -22.0 )
        return 0.125;
    else if (lat >= -62.0 )
        return 0.25;
    else if (lat >= -76.0 )
        return 0.5;
    else if (lat >= -83.0 )
        return 1.0;
    else if (lat >= -86.0 )
        return 2.0;
    else if (lat >= -88.0 )
          return 4.0;
    else if (lat >= -89.0 )
        return 8.0;
    else
        return 360.0;
}

wxString FGITile::FormatLon(wxDouble lon)
{
    if (lon < 0)
        return wxString::Format(wxT("w%03d"), int(0 - lon));
    else
        return wxString::Format(wxT("e%03d"), int(lon));
}

wxString FGITile::FormatLat(wxDouble lat)
{
    if (lat < 0)
        return wxString::Format(wxT("s%02d"), int(0 - lat));
    else
        return wxString::Format(wxT("n%02d"), int(lat));
}

wxString FGITile::GetNameFromDeg(wxDouble lon, wxDouble lat)
{
    wxDouble lon_floor = floor(lon);
    wxDouble lat_floor = floor(lat);
    wxDouble span = FGITile::BucketSpan(lat);

    wxInt32 x;
    if (span < FGITile::Epsilon)
    {
        lon = 0;
        x = 0;
    } else if (span <= 1.0) {
        x = int((lon - lon_floor) / span);
    } else {
        if (lon >= 0) {
            lon = int(int(lon / span) * span);
        } else {
            lon = int(int((lon + 1) / span) * span - span);
            if (lon < -180) {
                lon = -180;
            }
        }
        x = 0;
    }

    wxInt32 y;
    y = int((lat - lat_floor) * 8);

    wxDouble index = 0;
    index += (int(lon_floor) + 180) << 14;
    index += (int(lat_floor) + 90) << 6;
    index += y << 3;
    index += x;

    return wxString::Format(wxT("%d"), int(index));
}

wxString FGITile::GetPathFromDeg(const wxDouble lon, const wxDouble lat)
{
    wxDouble lon_floor = floor(lon);
    wxDouble lat_floor = floor(lat);
    wxDouble lon_chunk = floor(lon / 10) * 10;
    wxDouble lat_chunk = floor(lat / 10) * 10;

    return FGITile::FormatLon(lon_chunk) + FGITile::FormatLat(lat_chunk) + wxFileName::GetPathSeparator() + FGITile::FormatLon(lon_floor) + FGITile::FormatLat(lat_floor);
}

wxString FGITile::GetName()
{
    return Name;
}

wxString FGITile::GetPath()
{
    return Path;
}

